@@ -1,5 +1,6 @@
 use crate::config::application::Application;
 use crate::config::key::parse_key;
+use crate::config::key_action::KeyAction;
 use evdev::Key;
 use serde::de::{value, Error, MapAccess, Visitor};
 use serde::{Deserialize, Deserializer};
@@ -11,18 +12,18 @@ use std::fmt;
 pub struct Modmap {
     pub name: String,
     #[serde(deserialize_with = "modmap_remap")]
-    pub remap: HashMap<Key, Key>,
+    pub remap: HashMap<Key, KeyAction>,
     pub application: Option<Application>,
 }
 
-fn modmap_remap<'de, D>(deserializer: D) -> Result<HashMap<Key, Key>, D::Error>
+fn modmap_remap<'de, D>(deserializer: D) -> Result<HashMap<Key, KeyAction>, D::Error>
 where
     D: Deserializer<'de>,
 {
     struct ModmapRemap;
 
     impl<'de> Visitor<'de> for ModmapRemap {
-        type Value = HashMap<Key, Key>;
+        type Value = HashMap<Key, KeyAction>;
 
         fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
             formatter.write_str("map from string to string")
@@ -32,13 +33,12 @@ where
         where
             M: MapAccess<'de>,
         {
-            let remap: HashMap<String, String> = Deserialize::deserialize(value::MapAccessDeserializer::new(map))?;
+            let remap: HashMap<String, KeyAction> = Deserialize::deserialize(value::MapAccessDeserializer::new(map))?;
             let mut modmap = HashMap::new();
 
-            for (from, to) in remap.iter() {
+            for (from, to) in remap.into_iter() {
                 let from_key = parse_key(&from).map_err(M::Error::custom)?;
-                let to_key = parse_key(&to).map_err(M::Error::custom)?;
-                modmap.insert(from_key, to_key);
+                modmap.insert(from_key, to);
             }
 
             Ok(modmap)
